""" Helper class for working with alignment type analyses.
"""
import numpy as np

from ont_fast5_api.analysis_tools.base_tool import BaseTool
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.segmentation import SegmentationTools
from ont_fast5_api.fast5_read import Fast5Read


class AlignmentTools(BaseTool):
    """ Provides helper methods specific to alignment analyses.
    """
    
    def __init__(self, source, mode='r', group_name=None, meta=None, config=None):
        """ Create a new alignment tools object.
        
        :param source: Either an open Fast5File object, or a filename
            of a fast5 file.
        :param mode: The open mode (r or r+). Only if a filename is used
            for the source argument.
        :param group_name: The specific alignment analysis instance
            you are interested in.
        :param meta: Metadata for a new alignment analysis.
        :param config: Configuration data for a new alignment analysis.
        
        To create a new alignment analysis, provide a group name that
        does not already exist, and an optional dictionary with the metadata.
        The following fields are recommended, as a minimum:
            
            * name - The name of the basecall software used.
            * time_stamp - The time at which the analysis was performed.
        
        If the group name already exists, the "meta" parameter is ignored. If
        the specified group has a "component" attribute, and its value is not
        "alignment", an exception will be thrown.
        """
        if isinstance(source, Fast5Read):
            self.handle = source
            self.close_handle_when_done = False
        elif isinstance(source, str):
            self.handle = Fast5File(source, mode)
            self.close_handle_when_done = True
        else:
            raise Exception('Unrecognized type for argument "source".')
        if group_name is None:
            group_name = self.handle.get_latest_analysis('Alignment')
            if group_name is None:
                raise Exception('No Alignment analysis group found in file.')
        self.group_name = group_name
        attrs = self.handle.get_analysis_attributes(group_name)
        if attrs is None:
            if meta is None:
                meta = {}
            self.handle.add_analysis('alignment', group_name, meta, config)
            attrs = self.handle.get_analysis_attributes(group_name)
        if ('component' in attrs
                and attrs['component'] not in ['alignment',
                                               'calibration_strand']):
            self.close()
            raise Exception('Analysis does not appear to be an alignment component.')

    def get_results(self):
        """ Get details about the alignments that have been performed.

        :return: A dict of dicts.

        The keys of the top level are 'template', 'complement' and '2d'.
        Each of these dicts contains the following fields:

            * status: Can be 'no data', 'no match found', or 'match found'.
            * direction: Can be 'forward', 'reverse'.
            * ref_name: Name of reference.
            * ref_span: Section of reference aligned to, as a tuple (start, end).
            * seq_span: Section of the called sequence that aligned, as a tuple (start, end).
            * seq_len: Total length of the called sequence.
            * num_aligned: Number of bases that aligned to bases in the reference.
            * num_correct: Number of aligned bases that match the reference.
            * num_deletions: Number of bases in the aligned section of the
                reference that are not aligned to bases in the called sequence.
            * num_insertions: Number of bases in the aligned section of the called
                sequence that are not aligned to bases in the reference.
            * identity: The fraction of aligned bases that are correct (num_correct /
                num_aligned).
            * accuracy: The overall basecall accuracy, according to the alignment.
                (num_correct / (num_aligned + num_deletions + num_insertions)).
        
        Note that if the status field is not 'match found', then all the other
        fields will be absent.
        """
        summary = self.handle.get_summary_data(self.group_name)
        results = {'template': {'status': 'no data'},
                   'complement': {'status': 'no data'},
                   '2d': {'status': 'no data'}}
        if 'genome_mapping_template' in summary:
            results['template'] = self._get_results(summary['genome_mapping_template'])
        if 'genome_mapping_complement' in summary:
            results['complement'] = self._get_results(summary['genome_mapping_complement'])
        if 'genome_mapping_2d' in summary:
            results['2d'] = self._get_results(summary['genome_mapping_2d'])
        return results

    def get_alignment_data(self, section):
        """ Get the alignment SAM and Fasta, if present.
        
        :param section: Can be 'template', 'complement', or '2d'.
        :return: A tuple containing the SAM and the section of the reference
            aligned to (both as strings). Returns None if no alignment is
            present for that section.
        """
        subgroup = '{}/Aligned_{}'.format(self.group_name, section)
        sam = self.handle.get_analysis_dataset(subgroup, 'SAM')
        fasta = self.handle.get_analysis_dataset(subgroup, 'Fasta')
        if sam is None or fasta is None:
            return None
        sequence = fasta.split('\n')[1]
        return sam, sequence

    def add_alignment_data(self, section, sam, sequence):
        """ Add the SAM and Fasta alignment data for a section.
        
        :param section: Can be 'template', 'complement', or '2d'.
        :param sam: A string containing the SAM contents.
        :param sequence: A string containing the section of the
            reference the basecall aligned to.
        """
        subgroup = 'Aligned_{}'.format(section)
        if not subgroup in self.handle.handle['Analyses/{}'.format(self.group_name)]:
            self.handle.add_analysis_subgroup(self.group_name, subgroup)
        sam_arr = np.array(sam, dtype=str)
        self.handle.add_analysis_dataset('{}/{}'.format(self.group_name, subgroup), 'SAM', sam_arr)
        fasta_arr = np.array('>{}\n{}\n'.format(section, sequence), dtype=str)
        self.handle.add_analysis_dataset('{}/{}'.format(self.group_name, subgroup), 'Fasta', fasta_arr)

    def calculate_speed(self, section, alignment_results=None):
        """ Calculate speed using alignment information.

        :param section: The section (template or complement) we're calculating
            speed for.
        :param alignment_results: Optional dictionary of the alignment summary,
            so that speed can be calculated without having to write the summary
            out to the fast5 file first.
        :return: Speed in bases per second or zero if the speed could not be
            calculated.

        The only reliable way we have of finding out how many bases have gone through the pore is by
        looking at how much of the reference the sequence aligned to. This takes that information and
        uses it to calculate speed in reference-bases-per-second.
        """
        speed = 0.0
        if alignment_results:
            results = self._get_results(alignment_results)
        else:
            results = self.get_results()[section]
        if results['status'] != 'match found':
            return 0.0
        ref_span = results['ref_span']
        ref_len = ref_span[1] - ref_span[0]
        seq_span = results['seq_span']
        seq_len = seq_span[1] - seq_span[0]
        total_len = results['seq_len']

        sample_rate = self.handle.get_channel_info()['sampling_rate']

        # We need the duration from the segmentation results
        chain = self.handle.get_chain(self.group_name)
        if chain is not None:
            segmentation_group = dict(chain).get('segmentation')
        else:
            segmentation_group = None
        duration = 0
        if segmentation_group is not None:
            with SegmentationTools(self.handle, group_name=segmentation_group) as seg:
                summary = seg.get_results()
                if summary is not None:
                    duration = summary['duration_{}'.format(section)]
        if duration == 0:
            return 0.0

        normalized_duration = duration * seq_len / float(total_len)
        speed = sample_rate * ref_len / normalized_duration
        return speed

    ##########################
    #
    #  Private methods below
    #
    ##########################

    def _get_results(self, summary):
        results = {'status': 'no data'}
        ref_name = summary['genome']
        if ref_name == 'no_match':
            results['status'] = 'no match found'
            return results
        results['status'] = 'match found'
        results['direction'] = 'forward'
        if ref_name.endswith('_rc'):
            ref_name = ref_name[:-3]
            results['direction'] = 'reverse'
        results['ref_name'] = ref_name
        results['ref_span'] = (summary['genome_start'], summary['genome_end'])
        results['seq_span'] = (summary['strand_start'], summary['strand_end'])
        results['seq_len'] = summary['num_events']
        results.update({key: summary[key] for key in ['num_aligned', 'num_correct', 'num_insertions',
                                                      'num_deletions', 'identity', 'accuracy']})
        return results



from configparser import ConfigParser
from typing import Dict, Any, Union

import h5py
import numpy as np
import warnings
from collections import deque

from ont_fast5_api.compression_settings import VBZ, raise_missing_vbz_error_read, raise_missing_vbz_error_write
from ont_fast5_api.data_sanitisation import _sanitize_data_for_reading, _sanitize_data_for_writing, _clean
from ont_fast5_api.static_data import LEGACY_COMPONENT_NAMES


class AbstractFast5:
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()
        return False

    def get_reads(self):
        raise NotImplementedError()

    def get_read_ids(self):
        raise NotImplementedError()

    def get_read(self, read_id):
        raise NotImplementedError()

    @property
    def _is_open(self):
        return repr(self.handle) != u'<Closed HDF5 file>'

    def assert_open(self):
        if not self._is_open:
            raise IOError("Fast5 file is not open: {}".format(self.filename))

    def assert_writeable(self):
        self.assert_open()
        if self.mode == 'r':
            raise IOError("Fast5 file is in read-only mode '{}' {}".format(self.mode, self.filename))

    def close(self):
        """ Closes the object.
        """
        if self._is_open:
            self.mode = None
            if self.handle:
                self.handle.close()
                self.handle = None
            self.filename = None
            self.status = None


class Fast5Read(AbstractFast5):
    global_key = "" # This is for single Fast5File compatibility

    def __init__(self, parent, read_id):
        self.parent = parent
        self.mode = parent.mode
        self.filename = parent.filename
        self.handle = parent.handle["read_" + read_id]
        self.read_id = read_id


    def get_run_id(self):
        DeprecationWarning("'read.get_run_id()' will be deprecated. Use the property 'read.run_id' instead")
        return self.run_id

    def get_read_id(self):
        DeprecationWarning("'read.get_read_id()' will be deprecated. Use the property 'read.read_id' instead")
        return self.read_id

    @property
    def raw_dataset_name(self):
        return self.raw_dataset_group_name + "/Signal"

    @property
    def raw_dataset_group_name(self):
        return 'Raw'

    @property
    def raw_compression_filters(self):
        return self.handle[self.raw_dataset_name]._filters

    @property
    def run_id(self):
        return self.handle[self.global_key + 'tracking_id'].attrs['run_id']

    def add_raw_data(self, data, attrs=None, compression=VBZ):
        """ Add raw data for a read.

        :param data: The raw data DAQ values (16 bit integers).

        The read must already exist in the file. It must not already
        have raw data.
        """
        self.assert_writeable()
        if self.raw_dataset_group_name not in self.handle:
            self.handle.create_group(self.raw_dataset_group_name)
        if self.raw_dataset_name in self.handle:
            msg = "Fast5 file already has raw data for read '{}' in: {}"
            raise KeyError(msg.format(self.read_id, self.filename))
        try:
            self.handle[self.raw_dataset_group_name].create_dataset('Signal', data=data, dtype='i2',
                                                                    **vars(compression))
        except ValueError as e:
            raise_missing_vbz_error_read(e)

        if attrs is not None:
            self._add_attributes(self.raw_dataset_group_name, attrs, clear=True)

    def add_channel_info(self, attrs, clear=False):
        """ Add channel info data to the channel_id group.

        :param data: A dictionary of key/value pairs. Keys must be strings.
            Values can be strings or numeric values.
        :param clear: If set, any existing channel info data will be removed.
        """
        self.assert_writeable()
        if 'channel_id' not in self.handle:
            self.handle.create_group('channel_id')
        self._add_attributes('channel_id', attrs, clear)

    def add_tracking_id(self, attrs, clear=False):
        self.assert_writeable()
        if 'tracking_id' not in self.handle:
            self.handle.create_group("tracking_id")
        self.set_tracking_id(attrs, clear)

    def add_analysis(self, component, group_name, attrs, config=None):
        """ Add a new analysis group to the file.

        :param component: The component name.
        :param group_name: The name to use for the group. Must not already
            exist in the file e.g. 'Test_000'.
        :param attrs: A dictionary containing the key-value pairs to
            put in the analysis group as attributes. Keys must be strings,
            and values must be strings or numeric types.
        :param config: A dictionary of dictionaries. The top level keys
            should be the name of analysis steps, and should contain
            key value pairs for analysis parameters used.
        """
        self.assert_writeable()
        if "Analyses" not in self.handle:
            self.handle.create_group("Analyses")
        group = 'Analyses/{}'.format(group_name)
        cfg_group = '{}/Configuration'.format(group)
        group_attrs = attrs.copy()
        group_attrs['component'] = component
        self._add_group(group, group_attrs)
        self.handle[group].create_group('Summary')
        self.handle[group].create_group('Configuration')
        if config is not None:
            self._add_attribute_tree(cfg_group, config)

    def get_raw_data(self, read_number=None, start=None, end=None, scale=False):
        if read_number:
            warnings.warn("Read number is not used for MultiReadFast5")
        return self._load_raw(self.raw_dataset_name, start, end, scale)

    def list_analyses(self, component=None):
        """ Provides a list of all analyses groups.

        :param component: Optional component name. If provided, only
            analyses of that component will be returned.
        :returns: A list of component-name/group-name pairs (tuples).
        """
        self.assert_open()
        analyses = []
        if 'Analyses' not in self.handle:
            return analyses
        ana_groups = self.handle['Analyses'].keys()
        for group_name in ana_groups:
            group_attrs = self.handle['Analyses/{}'.format(group_name)].attrs
            if 'component' in group_attrs:
                comp = _clean(group_attrs['component'])
            elif group_name[:-4] in LEGACY_COMPONENT_NAMES:
                comp = LEGACY_COMPONENT_NAMES[group_name[:-4]]
            else:
                # We don't know anything about this component!
                comp = None
            if comp is not None and (component is None or comp == component):
                analyses.append((comp, group_name))
        return analyses

    def get_latest_analysis(self, group_base, increment=False):
        """ Return the latest analysis group with the specified
            base name.

        :param group_base: The name of the analysis group, excluding
            the _### at the end e.g. 'Alignment' rather than 'Alignment_000'
        :param increment: If set, this will find the name of the latest
            analysis group, increment the index by 1, and return
            the results. If there aren't any, it will give an index
            of 000.
        :returns: The group name with the highest index, or None if
            there are no analyses with the specified base name.
        """
        self.assert_open()
        all_analyses = self.list_analyses()
        selected = []
        for analysis in all_analyses:
            if analysis[1][:-4] == group_base:
                selected.append(analysis[1])
        if len(selected) == 0:
            result = None
            if increment:
                result = '{}_000'.format(group_base)
            return result
        result = sorted(selected)[-1]
        if increment:
            count = int(result[-3:]) + 1
            result = '{}_{}'.format(group_base, str(count).zfill(3))
        return result

    def add_chain(self, group_name, component_map):
        """
        Adds the component chain to ``group_name`` in the fast5.
        These are added as attributes to the group.

        :param group_name: The group name you wish to add chaining data to,
            e.g. ``Test_000``
        :param component_map: The set of components and corresponding
            group names or group paths that contribute data to the analysis.
            If group names are provided, these will be converted into group
            paths.

            If ``Test_000`` uses data from the results of
            ``first_component`` stored at ``Analyses/First_000/``
            the component_map could be ``{'first_component': 'First_000'}`` or
            ``{'first_component': 'Analyses/First_000'}``.

        """
        self.assert_writeable()

        for component, path in component_map.items():
            if not path.startswith('Analyses/'):
                path = 'Analyses/{}'.format(path)
            component_map[component] = path

        self.add_analysis_attributes(group_name, component_map)

    def get_chain(self, group_name):
        """ Provides the component and group names for an analysis chain.

        :param group_name: The group name of the last step in the analysis
            chain e.g. 'Basecall_1D_000'
        :returns: A list of component-name/group-name pairs (tuples). This
            will include each component of the chain, in order.
        """
        self.assert_open()
        endgroup = 'Analyses/{}'.format(group_name)
        attr = self.handle[endgroup].attrs
        if 'component' in attr:
            component = attr['component']
        else:
            component = LEGACY_COMPONENT_NAMES[group_name[:-4]]
        chain = deque()
        chain.append((component, group_name))
        groups_to_check = deque()
        groups_to_check.append(endgroup)
        while len(groups_to_check) > 0:
            group = groups_to_check.popleft()
            attr = self.handle[group].attrs
            for key, value in attr.items():
                if str(value).startswith('Analyses/'):
                    chain_entry = (key, value[9:])
                    # We need to maintain the order of the components, so
                    # we'll move any we see again to the end of the chain.
                    if chain_entry in chain:
                        chain.remove(chain_entry)
                    chain.append(chain_entry)
                    groups_to_check.append(value)
        return list(chain)

    def get_tracking_id(self):
        """ Returns a dictionary of tracking-id key/value pairs.
        """
        self.assert_open()
        tracking = self.handle[self.global_key + 'tracking_id'].attrs.items()
        tracking = {key: _clean(value) for key, value in tracking}
        return tracking

    def set_tracking_id(self, data, clear=False):
        """ Add tracking-id data to the tracking_id group.

        :param data: A dictionary of key/value pairs. Keys must be strings.
            Values can be strings or numeric values.
        :param clear: If set, any existing tracking-id data will be removed. 
        """
        self.assert_writeable()
        self._add_attributes(self.global_key + 'tracking_id', data, clear)
        return

    def get_channel_info(self):
        """ Returns a dictionary of channel information key/value pairs.
        """
        self.assert_open()
        channel_info = self.handle[self.global_key + 'channel_id'].attrs.items()
        channel_info = {key: _clean(value) for key, value in channel_info}
        channel_info['channel_number'] = int(channel_info['channel_number'])
        return channel_info

    @property
    def has_context_tags(self):
        return 'context_tags' in self.handle

    def get_context_tags(self):
        """ Returns a dictionary of context tag key/value pairs.
        """
        self.assert_open()
        if self.has_context_tags:
            tags = self.handle[self.global_key + 'context_tags'].attrs.items()
            return {key: _clean(value) for key, value in tags}
        return {}

    def add_context_tags(self, data, clear=False):
        """ Replaces any existing context tag data with the provided values.

        :param data: A dictionary of key/value pairs. Keys must be strings.
            Values can be strings or numeric values.
        :param clear: If set, any existing context tag data will be removed. 
        """
        self.assert_writeable()
        if self.has_context_tags:
            self._add_attributes(self.global_key + 'context_tags', data, clear)
        else:
            self._add_group(self.global_key + 'context_tags', data)

    def add_read(self, read_number, read_id, start_time, duration, mux, median_before):
        raise NotImplementedError("Cannot add_existing_read() to a Fast5Read(). "
                                  "Use MultiFast5File.create_empty_read() instead")

    def add_log(self, group_name, field_name, log_string):
        """ Add a log the file.
        :param group_name: Global group name. Cave: No 'Analyses' is prepended.
        :param field_name: A field {group_name}/{field_name} will be written.
        :param log_string: content to be written.
        """
        self.assert_writeable()
        if group_name not in self.handle:
            self._add_group(group_name, {})
        sanitized_data = _sanitize_data_for_writing(log_string)
        self.handle[group_name].create_dataset(field_name, data=sanitized_data)

    def set_summary_data(self, group_name, section_name, data):
        """ Set the summary data for an analysis group.

        :param group_name: The name of the analysis group.
        :param section_name: The analysis step. This will become a
            subfolder in the Summary section.
        :param data: A dictionary containing keys which are the summary
            fields, and values which are the summary values.
        """
        self.assert_writeable()
        group = 'Analyses/{}/Summary/{}'.format(group_name, section_name)
        self._add_group(group, data)

    def set_analysis_config(self, group_name, config):
        """ Set Configuration data for analysis group.  The ``config`` can
            be passed as a dict of dicts e.g. ``{'section': {'key': 'value'}}``
            or can be passed directly as a ConfigParser object.

        :param group_name: The name of the analysis group e.g. Example_000
        :param config: Representation of configuration as ConfigParser or
            dict of dicts.
        :raises KeyError: if group_name does not exist in fast5 or
            fast5 is not open for writing
        :raises TypeError: if ``config`` is not ConfigParser or dict obj.
        """
        self.assert_writeable()
        if 'Analyses/{}'.format(group_name) not in self.handle:
            msg = 'Dataset cannot be added to non-existent group: Analyses/{} in {}'
            raise KeyError(msg.format(group_name, self.filename))
        if isinstance(config, ConfigParser):
            config_dict = {}
            for section in config.sections():
                config_dict[section] = {k: v for k, v in config.items(section)}
        elif isinstance(config, dict):
            config_dict = config
        else:
            raise TypeError('config must be a ConfigParser or dict not {}'.format(type(config)))

        config_path = 'Analyses/{}/Configuration'.format(group_name)
        self._add_attribute_tree(config_path, config_dict)

    def get_analysis_config(self, group_name):
        """ Gets any config data saved for the analysis.

        :param group_name: The name of the analysis group.
        :returns: A dictionary of dictionaries. Each key represents
            an analysis step. Each value is a dictionary containing the
            analysis parameters as key/value pairs. Returns None if no
            configuration exists for the analysis.
        """
        self.assert_open()
        group = 'Analyses/{}/Configuration'.format(group_name)
        config = None
        if group in self.handle:
            config = self._parse_attribute_tree(group)
        return config

    def get_summary_data(self, group_name):
        """ Get the summary data for an analysis group.

        :param group_name: The name of the analysis group to pull summary
            data for.
        :returns: A dictionary whose keys are analysis steps, and whose
            values are dictionaries of key/value pairs for the results of
            that step.
        """
        self.assert_open()
        group = 'Analyses/{}/Summary'.format(group_name)
        summary = None
        if group in self.handle:
            summary = self._parse_attribute_tree(group)
        return summary

    def add_analysis_subgroup(self, group_name, subgroup_name, attrs=None):
        """ Add a new subgroup to an existing analysis group.

        :param group_name: The name of the analysis group you are adding to.
        :param subgroup_name: The name of the new subgroup.
        :param attrs: A dictionary representing the attributes to assign to
            the subgroup.

        The new subgroup must not already exist.

        The subgroup name can be a nested name, such as "Template/Data". This
        will create the "Template" subgroup (if it does not exist), and the
        "Data" subgroup below it.
        """
        self.assert_writeable()
        group = 'Analyses/{}/{}'.format(group_name, subgroup_name)
        self._add_group(group, attrs)

    def add_analysis_attributes(self, group_name, attrs, clear=False):
        """ Add attributes on the group or dataset specified.

        :param group_name: The name of the group (or dataset).
        :param attrs: A dictionary representing the attributes to add.
        :param clear: If set, any existing attributes will be cleared.

        The specified group name can be any existing path (relative to the
        "Analyses" group. It can be a group or a dataset.
        """
        self.assert_writeable()
        group = 'Analyses/{}'.format(group_name)
        self._add_attributes(group, attrs, clear)

    def get_analysis_attributes(self, group_name):
        """ Returns the attributes for the specified group or dataset.

        :param group_name: The path of the group or dataset, relative to the
            "Analyses" group.
        :returns: A dictionary representing the attributes (if any).
        """
        self.assert_open()
        group = 'Analyses/{}'.format(group_name)
        attr = None
        if group in self.handle:
            attr = self.handle[group].attrs.items()
            attr = {key: _clean(value) for key, value in attr}
        return attr

    def add_analysis_dataset(self, group_name, dataset_name, data, attrs=None):
        """ Add a dataset to the specified group.

        :param group_name: The path of the group the dataset will be added to,
            relative to the "Analyses" group.
        :param dataset_name: The name of the new dataset.
        :param data: A numpy array representing the data to be written.
        :param attrs: A dictionary of attributes to be added to the dataset.
        :raises KeyError: If dataset is being added to non-existant group or
            if file is not open for writing.
        """
        self.assert_writeable()
        group_path = 'Analyses/{}'.format(group_name)
        if group_path not in self.handle:
            msg = 'Dataset cannot be added to non-existent group: Analyses/{} in {}'
            raise KeyError(msg.format(group_name, self.filename))

        sanitized_data = _sanitize_data_for_writing(data)
        if np.shape(sanitized_data) == ():  # We can't compress scalar datasets
            self.handle[group_path].create_dataset(dataset_name,
                                                   data=sanitized_data)
        else:
            self.handle[group_path].create_dataset(dataset_name,
                                                   data=sanitized_data,
                                                   compression='gzip')
        if attrs is not None:
            path = '{}/{}'.format(group_path, dataset_name)
            self._add_attributes(path, attrs)

    def get_analysis_dataset(self, group_name, dataset_name, skip_decoding=False):
        """ Return the specified dataset, as a numpy array.

        :param group_name: The path of the group containing the dataset,
            relative to the "Analyses" group.
        :param dataset_name: The name of the dataset.
        :param skip_decoding: If True, this will directly return the h5py dataset.
            If False, it will load the data into a Numpy array and decode any byte-strings
        :return: A numpy array containing the data. Returns None if the dataset
            does not exist.
        """
        self.assert_open()
        dataset_name = 'Analyses/{}/{}'.format(group_name, dataset_name)
        data = None
        if dataset_name in self.handle:
            data = self.handle[dataset_name]
            if not skip_decoding:
                data = _sanitize_data_for_reading(data)
        return data

    @staticmethod
    def read_summary_data(fname, component):
        raise NotImplementedError("read_summary_data() is not implemented for MultiFast5 reads")

    @staticmethod
    def update_legacy_file(fname):
        raise NotImplementedError("Cannot update legacy file for a MultiFast5Read")

    def _load_raw(self, dataset_name, start, end, scale):
        try:
            raw = self.handle[dataset_name][start:end]
        except OSError as err:
            raise_missing_vbz_error_write(err)
        if scale:
            channel_info = self.handle[self.global_key + 'channel_id'].attrs
            digi = channel_info['digitisation']
            parange = channel_info['range']
            offset = channel_info['offset']
            scaling = parange / digi
            # python slice syntax allows None, https://docs.python.org/3/library/functions.html#slice
            raw = np.array(scaling * (raw + offset), dtype=np.float32)
        return raw

    def _add_group(self, group, attrs):
        """
        :param group: group_name
        :param attrs:
        :return:
        """
        self.handle.create_group(group)
        if attrs is not None:
            self._add_attributes(group, attrs)

    def _add_attributes(self, path, attrs, clear=False):
        path_grp = self.handle[path]
        path_attr = path_grp.attrs
        if clear:
            for key in path_attr.keys():
                del path_attr[key]

        copy_attributes(attrs, path_grp)

    def _add_attribute_tree(self, group, config):
        for folder, data in config.items():
            path = '{}/{}'.format(group, folder)
            self._add_group(path, data)

    def _parse_attribute_tree(self, group):
        data = {}
        folders = self.handle[group].keys()
        for folder in folders:
            path = '{}/{}'.format(group, folder)
            attr = self.handle[path].attrs.items()
            data[folder] = {key: _clean(value) for key, value in attr}
        return data


def copy_attributes(input_attrs: Union[h5py.AttributeManager, Dict[Any, Any]], output_group: h5py.Group):
    """ Copy the members of a given h5py.AttributeManager into an output h5py.Group"""
    if isinstance(input_attrs, h5py.AttributeManager):
        # Ensure h5py data types (e.g. enumerations) are retained during copy
        for name, data in input_attrs.items():
            dtype = input_attrs.get_id(name).dtype
            output_group.attrs.create(name=name, data=data, dtype=dtype)
    else:
        # Simple dictionary copy into the output group attributes
        for name, data in input_attrs.items():
            output_group.attrs[name] = data


import os
import unittest

import numpy

from ont_fast5_api.compression_settings import GZIP, VBZ_ALPHA
from ont_fast5_api.conversion_tools.check_file_compression import check_read_compression, check_compression
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.fast5_interface import get_fast5_file
from ont_fast5_api.multi_fast5 import MultiFast5File
from test.helpers import test_data


class TestCheckCompression(unittest.TestCase):

    def test_check_read_compression_single_read(self):
        with get_fast5_file(os.path.join(test_data, 'single_reads', 'read0.fast5'), 'r') as f5:
            for read in f5.get_reads():
                compression = check_read_compression(read)
                self.assertEqual(compression, GZIP)

    def test_check_read_compression_multi_read(self):
        with get_fast5_file(os.path.join(test_data, 'multi_read', 'batch_0.fast5'), 'r') as f5:
            for read in f5.get_reads():
                compression = check_read_compression(read)
                self.assertEqual(compression, GZIP)

    def test_check_read_compression_vbz(self):
        with get_fast5_file(os.path.join(test_data, 'vbz_reads', 'vbz_reads.fast5'), 'r') as f5:
            for read in f5.get_reads():
                compression = check_read_compression(read)
                self.assertEqual(compression, VBZ_ALPHA)

    def test_check_single_read_folder(self):
        input_folder = os.path.join(test_data, 'single_reads')
        compression_results = list(check_compression(input_folder, recursive=False,
                                                     follow_symlinks=False, check_all_reads=False))

        ## expected
        expected_results = []
        for input_file in os.listdir(input_folder):
            input_path = os.path.join(input_folder, input_file)
            with Fast5File(input_path, 'r') as f5:
                expected_results.append((GZIP, f5.read_id, input_path))

        self.assertTrue(numpy.array_equal(expected_results, compression_results))

    def test_check_multi_read(self):
        input_folder = os.path.join(test_data, 'vbz_reads')
        ## expected results
        expected_results = []
        for input_file in os.listdir(input_folder):
            input_path = os.path.join(input_folder, input_file)
            with MultiFast5File(input_path, 'r') as f5:
                for read in f5.get_reads():
                    expected_results.append((VBZ_ALPHA, read.read_id, input_path))

        # Test check all reads True
        compression_results = list(check_compression(input_folder, recursive=False, follow_symlinks=False,
                                                     check_all_reads=True))
        self.assertTrue(numpy.array_equal(expected_results, compression_results))

        ## check one read only
        compression_results = list(check_compression(input_folder, recursive=False, follow_symlinks=False,
                                                     check_all_reads=False))
        self.assertTrue(len(compression_results) == len(os.listdir(input_folder)))
        self.assertTrue(compression_results[0] in expected_results)



from unittest.mock import patch

import os
import h5py
import numpy

from ont_fast5_api.conversion_tools.multi_to_single_fast5 import convert_multi_to_single, try_multi_to_single_conversion
from ont_fast5_api.conversion_tools.single_to_multi_fast5 import batch_convert_single_to_multi, get_fast5_file_list, \
    create_multi_read_file
from ont_fast5_api.multi_fast5 import MultiFast5File
from ont_fast5_api.fast5_file import Fast5FileTypeError, Fast5File
from test.helpers import TestFast5ApiHelper, test_data, disable_logging


class TestFast5Converter(TestFast5ApiHelper):

    @patch('ont_fast5_api.conversion_tools.single_to_multi_fast5.get_progress_bar')
    def test_single_to_multi(self, mock_pbar):
        input_folder = os.path.join(test_data, "single_reads")
        batch_size = 3
        file_count = len(os.listdir(input_folder))
        batch_convert_single_to_multi(input_folder, self.save_path, filename_base="batch", batch_size=batch_size,
                                      threads=1, recursive=False, follow_symlinks=False, target_compression=None)

        expected_output_reads = {"filename_mapping.txt": 0,
                                 "batch_0.fast5": batch_size,
                                 "batch_1.fast5": file_count % batch_size}
        self.assertEqual(sorted(os.listdir(self.save_path)), sorted(list(expected_output_reads.keys())))
        for file, read_count in expected_output_reads.items():
            if read_count > 0:
                with h5py.File(os.path.join(self.save_path, file), 'r') as f5:
                    self.assertEqual(len(f5), read_count)

    def test_multi_to_single(self):
        input_file = os.path.join(test_data, "multi_read", "batch_0.fast5")
        with MultiFast5File(input_file, 'r') as f5:
            read_count = len(f5.handle)
            expected_files = sorted([os.path.join(self.save_path, "{}", i + '.fast5') for i in f5.get_read_ids()])

        subfolder = '0'
        convert_multi_to_single(input_file, self.save_path, subfolder)

        out_files = sorted(get_fast5_file_list(self.save_path, recursive=True, follow_symlinks=True))
        self.assertEqual(len(out_files), read_count)
        self.assertEqual(out_files, [f.format(subfolder) for f in expected_files])

    @disable_logging
    def test_single_to_multi_incorrect_types(self):
        input_files = [os.path.join(test_data, "multi_read", "batch_0.fast5")]
        with self.assertRaises(Fast5FileTypeError):
            create_multi_read_file(input_files, self.generate_temp_filename(), target_compression=None)

    def test_multi_to_single_incorrect_types(self):
        input_folder = os.path.join(test_data, "single_reads")
        input_file = os.path.join(input_folder, os.listdir(input_folder)[0])
        with self.assertRaises(Fast5FileTypeError):
            try_multi_to_single_conversion(input_file, self.save_path, subfolder='0')

    def test_add_read_to_multi(self):
        with Fast5File(os.path.join(test_data, "single_reads", "read0.fast5"), 'r') as single_fast5, \
                MultiFast5File(self.generate_temp_filename(), 'w') as multi_out:
            multi_out.add_existing_read(single_fast5)
            expected_raw = single_fast5.get_raw_data()
            actual_raw = multi_out.get_read(single_fast5.get_read_id()).get_raw_data()
            self.assertTrue(numpy.array_equal(actual_raw, expected_raw))



""" Helper class for getting information about a fast5 file.
"""
import os
import h5py

from packaging import version as packaging_version

# This unused import is included for backwards compatibilty and can be removed in future.
from ont_fast5_api.data_sanitisation import _clean

class ReadInfo(object):
    """ This object provides basic details about a read.
    """
    
    def __init__(self, read_number, read_id, start_time, duration,
                 mux=0, median_before=-1.0):
        """ Constructs an object describing a read.
        
        :param read_number: A read number, unique for the channel.
        :param read_id: A globally unique read id.
        :param start_time: The start time of the read (in samples).
        :param duration: The duration of the read (in samples).
        :param mux: The mux of the channel when the read occurred.
        :param median_before: The median current before the read.
        """
        self.read_number = read_number
        self.read_id = read_id
        self.has_raw_data = False
        self.duration = duration
        self.has_event_data = False
        self.event_data_count = 0
        self.start_time = start_time
        self.start_mux = mux
        self.median_before = median_before


class Fast5Info(object):
    """ This object provides some basic details about a read fast5 file.
    
    **Fields**
      * **valid:** Indicates whether the fast5 file is valid or not.
      * **version:** Indicates the version of the read fast5 file
        specification the file conforms to (if any).
      * **read_info:** A list of ReadInfo objects. One entry for each read.
      * **read_number_map:** A dictionary giving the index into the read_info
        list for each read number.
      * **read_id_map:** A dictionary giving the index into the read_info
        list for each read-id.
    """

    def __init__(self, fname):
        """ Constructs a status object from a file.

        :param fname: Filename of fast5 file to read status from.
        """
        self.valid = True
        self.channel = None
        self.read_info = []
        self.read_number_map = {}
        self.read_id_map = {}
        try:
            with h5py.File(fname, 'r') as handle:
                if 'file_version' in handle.attrs:
                    self.version = _clean(handle.attrs['file_version'])
                    minimum_valid_version = packaging_version.Version('0.6')
                    if packaging_version.parse(str(self.version)) \
                       < minimum_valid_version:
                        self.valid = False
                else:
                    self.valid = False
                    self.version = 0.0

                # Check for required groups.
                top_groups = handle.keys()
                if 'UniqueGlobalKey' in top_groups:
                    global_keys = handle['UniqueGlobalKey'].keys()
                if 'tracking_id' not in global_keys and not self._legacy_version():
                    self.valid = False
                if 'channel_id' not in global_keys:
                    self.valid = False

                self.channel = handle['UniqueGlobalKey/channel_id'].attrs.get('channel_number')
                if self.channel is None and self._legacy_version():
                    self.valid = False

                # Get the read information.
                if 'Raw' in top_groups:
                    reads = handle['Raw/Reads'].keys()
                    for read in reads:
                        read_group_name = 'Raw/Reads/{}'.format(read)
                        read_group = handle[read_group_name]
                        read_attrs = read_group.attrs
                        read_number = _clean(read_attrs['read_number'])
                        if 'read_id' in read_attrs:
                            read_id = _clean(read_attrs['read_id'])
                        else:
                            if not self._legacy_version():
                                self.valid = False
                            else:
                                read_id = os.path.basename(fname)
                        start_time = _clean(read_attrs['start_time'])
                        duration = _clean(read_attrs['duration'])
                        mux = _clean(read_attrs.get('start_mux',0))
                        median_before = _clean(read_attrs.get('median_before',-1.0))
                        read_info = ReadInfo(read_number, read_id, start_time, duration, mux, median_before)
                        if 'Signal' in read_group:
                            read_info.has_raw_data = True
                        elif self._legacy_version():
                            if 'Data' in read_group:
                                read_info.has_raw_data = True
                            else:
                                self.valid = False
                        self.read_info.append(read_info)
                        n = len(self.read_info) - 1
                        self.read_number_map[read_number] = n
                        self.read_id_map[read_id] = n
                else:
                    if not self._legacy_version():
                        self.valid = False
                analyses = sorted(handle['Analyses'].keys()) if 'Analyses' in handle else []
                for ana in analyses[::-1]:
                    if ana.startswith('EventDetection'):
                        reads_group_name = 'Analyses/{}/Reads'.format(ana)
                        if reads_group_name not in handle:
                            continue
                        reads = handle[reads_group_name].keys()
                        for read in reads:
                            read_group_name = '{}/{}'.format(reads_group_name, read)
                            read_group = handle[read_group_name]
                            read_attrs = read_group.attrs
                            read_number = _clean(read_attrs['read_number'])
                            if 'read_id' in read_attrs:
                                read_id = _clean(read_attrs['read_id'])
                            else:
                                if not self._legacy_version():
                                    self.valid = False
                                    continue
                                else:
                                    read_id = os.path.basename(fname)
                            start_time = _clean(read_attrs['start_time'])
                            duration = _clean(read_attrs['duration'])
                            mux = _clean(read_attrs.get('start_mux', 0))
                            median_before = _clean(read_attrs.get('median_before', -1.0))
                            read_info = ReadInfo(read_number, read_id, start_time, duration, mux, median_before)
                            if 'Events' in read_group:
                                read_info.has_event_data = True
                                read_info.event_data_count = len(read_group['Events'])
                            else:
                                read_info.has_event_data = False
                                read_info.event_data_count = 0
                            if read_number in self.read_number_map:
                                read_index = self.read_number_map[read_number]
                                self.read_info[read_index].has_event_data = read_info.has_event_data
                                self.read_info[read_index].event_data_count = read_info.event_data_count
                            else:
                                if not self._legacy_version():
                                    self.valid = False
                                self.read_info.append(read_info)
                                n = len(self.read_info) - 1
                                self.read_number_map[read_number] = n
                                self.read_id_map[read_id] = n
                        break
        except:
            self.valid = False
            raise

        if self._legacy_version():
            # There must be either raw data or event data (or both).
            if len(self.read_info) == 0:
                self.valid = False

    def _legacy_version(self):
        legacy_cutoff = packaging_version.Version("1.1")
        return packaging_version.parse(str(self.version)) < legacy_cutoff



import logging
import os
from argparse import ArgumentParser
from multiprocessing import Pool

from ont_fast5_api import __version__
from ont_fast5_api.compression_settings import COMPRESSION_MAP
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list, batcher, get_progress_bar
from ont_fast5_api.fast5_file import Fast5File, Fast5FileTypeError
from ont_fast5_api.multi_fast5 import MultiFast5File

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
exc_info = False


def batch_convert_single_to_multi(input_path, output_folder, filename_base, batch_size,
                                  threads, recursive, follow_symlinks, target_compression):
    pool = Pool(threads)
    file_list = get_fast5_file_list(input_path, recursive, follow_symlinks)
    pbar = get_progress_bar(int((len(file_list) + batch_size - 1) / batch_size))

    def update(result):
        output_file = result[1]
        with open(os.path.join(output_folder, "filename_mapping.txt"), 'a') as output_table:
            for filename in result[0]:
                output_table.write("{}\t{}\n".format(filename, output_file))
        pbar.update(pbar.currval + 1)

    results_array = []
    os.makedirs(output_folder, exist_ok=True)
    for batch_num, batch in enumerate(batcher(file_list, batch_size)):
        output_file = os.path.join(output_folder, "{}_{}.fast5".format(filename_base, batch_num))
        results_array.append(pool.apply_async(create_multi_read_file,
                                              args=(batch, output_file, target_compression),
                                              callback=update))

    pool.close()
    pool.join()
    pbar.finish()


def create_multi_read_file(input_files, output_file, target_compression):
    results = []
    os.makedirs(os.path.dirname(output_file), exist_ok=True)
    if os.path.exists(output_file):
        logger.info("FileExists - appending new reads to existing file: {}".format(output_file))
    try:
        with MultiFast5File(output_file, 'a') as multi_f5:
            for filename in input_files:
                try:
                    with Fast5File(filename, 'r') as f5_input:
                        read = f5_input.get_read(f5_input.read_id)
                        multi_f5.add_existing_read(read, target_compression=target_compression)
                    results.append(os.path.basename(filename))
                except Fast5FileTypeError as e:
                    logger.error("{}: Cannot input MultiRead files to single_to_multi: '{}'"
                                 "".format(e, filename), exc_info=exc_info)
                    raise
                except Exception as e:
                    logger.error("{}\n\tFailed to add single read file: '{}' to '{}'"
                                 "".format(e, filename, output_file), exc_info=exc_info)

    except Fast5FileTypeError:
        raise
    except Exception as e:
        logger.error("{}\n\tFailed to write to MultiRead file: {}"
                     "".format(e, output_file), exc_info=exc_info)
    return results, output_file


def main():
    parser = ArgumentParser("")
    parser.add_argument('-i', '--input_path', required=True,
                        help='Folder containing single read fast5 files')
    parser.add_argument('-s', '--save_path', required=True,
                        help="Folder to output multi read files to")
    parser.add_argument('-f', '--filename_base', default='batch', required=False,
                        help="Root of output filename, default='batch' -> 'batch_0.fast5'")
    parser.add_argument('-n', '--batch_size', type=int, default=4000, required=False,
                        help="Number of reads per multi-read file")
    parser.add_argument('-t', '--threads', type=int, default=1, required=False,
                        help="Number of threads to use")
    parser.add_argument('--recursive', action='store_true',
                        help="Search recursively through folders for single_read fast5 files")
    parser.add_argument('--ignore_symlinks', action='store_true',
                        help="Ignore symlinks when searching recursively for fast5 files")
    parser.add_argument('-c', '--compression', required=False, default=None,
                        choices=list(COMPRESSION_MAP.keys()) + [None], help="Target output compression type")
    parser.add_argument('-v', '--version', action='version', version=__version__)
    args = parser.parse_args()

    if args.compression is not None:
        args.compression = COMPRESSION_MAP[args.compression]

    batch_convert_single_to_multi(args.input_path,
                                  args.save_path,
                                  args.filename_base,
                                  args.batch_size,
                                  args.threads,
                                  args.recursive,
                                  follow_symlinks=not args.ignore_symlinks,
                                  target_compression=args.compression)


if __name__ == '__main__':
    main()



import os
import re
from setuptools import setup, find_packages

__pkg_name__ = 'ont_fast5_api'


def get_version():
    init_file = os.path.join(__pkg_name__, '__init__.py')
    with open(init_file, 'r') as init_fh:
        verstrline = init_fh.read()
    vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(vsre, verstrline, re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in '{}'".format(init_file))


with open('README.rst') as readme:
    documentation = readme.read()

installation_requirements = []
if 'IGNORE_INCLUDES' not in os.environ:
    installation_requirements = ['h5py>=3',
                                 'numpy>=1.16',
                                 'packaging',
                                 'progressbar33>=2.3.1',
                                 'setuptools']

setup(name=__pkg_name__.replace("_", "-"),
      author='Oxford Nanopore Technologies, Limited',
      description='Oxford Nanopore Technologies fast5 API software',
      license='MPL 2.0',
      long_description=documentation,
      version=get_version(),
      url='https://github.com/nanoporetech/{}'.format(__pkg_name__),
      install_requires=installation_requirements,
      packages=find_packages(),
      package_data={__pkg_name__: ['vbz_plugin/*.so', 'vbz_plugin/*.dylib', 'vbz_plugin/*.dll']},
      python_requires='>=3.7',
      entry_points={'console_scripts': [
          "multi_to_single_fast5={}.conversion_tools.multi_to_single_fast5:main".format(__pkg_name__),
          "single_to_multi_fast5={}.conversion_tools.single_to_multi_fast5:main".format(__pkg_name__),
          "fast5_subset={}.conversion_tools.fast5_subset:main".format(__pkg_name__),
          "compress_fast5={}.conversion_tools.compress_fast5:main".format(__pkg_name__),
          "check_compression={}.conversion_tools.check_file_compression:main".format(__pkg_name__),
          "demux_fast5={}.conversion_tools.demux_fast5:main".format(__pkg_name__),
      ]},
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
          'Natural Language :: English',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='fast5 nanopore')



import os
import numpy as np
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.alignment import AlignmentTools
from test.helpers import TestFast5ApiHelper


class TestAlignmentTools(TestFast5ApiHelper):

    def test_001_put_and_retrieve(self):
        fname = self.generate_temp_filename()
        summary_temp = {'genome': 'Lambda',
                        'genome_start': 100,
                        'genome_end': 200,
                        'strand_start': 1,
                        'strand_end': 101,
                        'num_events': 125,
                        'num_aligned': 92,
                        'num_correct': 87,
                        'num_insertions': 8,
                        'num_deletions': 8,
                        'identity': 0.9457,
                        'accuracy': 0.8056}
        summary_comp = {'genome': 'Lambda_rc',
                        'genome_start': 100,
                        'genome_end': 200,
                        'strand_start': 0,
                        'strand_end': 96,
                        'num_events': 120,
                        'num_aligned': 90,
                        'num_correct': 88,
                        'num_insertions': 6,
                        'num_deletions': 10,
                        'identity': 0.9778,
                        'accuracy': 0.8302}
        summary_2d = {'genome': 'Lambda',
                      'genome_start': 100,
                      'genome_end': 200,
                      'strand_start': 0,
                      'strand_end': 100,
                      'num_events': 125,
                      'num_aligned': 98,
                      'num_correct': 96,
                      'num_insertions': 4,
                      'num_deletions': 4,
                      'identity': 0.9796,
                      'accuracy': 0.9057}
        sam1 = 'Dummy string for template SAM.'
        sam2 = 'Dummy string for complement SAM.'
        sam3 = 'Dummy string for 2D SAM.'
        sequence1 = ''.join(np.random.choice(['A', 'C', 'G', 'T'], 100))
        bc = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
        sequence2 = ''.join([bc[letter] for letter in sequence1[::-1]])
        with Fast5File(fname, mode='w') as fh:
            fh.add_channel_info({'channel_number': 1,
                                 'sampling_rate': 4000,
                                 'digitisation': 8192,
                                 'range': 819.2,
                                 'offset': 0})
            fh.add_read(12, 'unique_snowflake', 12345, 4000, 0, 120.75)
            attrs = {'name': 'test', 'version': 0, 'time_stamp': 'just now', 'component': 'segmentation'}
            fh.add_analysis('segmentation', 'Segmentation_000', attrs)
            seg_data = {'has_template': 1,
                        'has_complement': 1,
                        'first_sample_template': 0,
                        'duration_template': 2000,
                        'first_sample_complement': 2000,
                        'duration_complement': 2000}
            fh.set_summary_data('Segmentation_000', 'segmentation', seg_data)
            attrs['component'] = 'alignment'
            attrs['segmentation'] = 'Analyses/Segmentation_000'
            fh.add_analysis('alignment', 'Alignment_000', attrs)
            fh.set_summary_data('Alignment_000', 'genome_mapping_template', summary_temp)
            fh.set_summary_data('Alignment_000', 'genome_mapping_complement', summary_comp)
            fh.set_summary_data('Alignment_000', 'genome_mapping_2d', summary_2d)
            with AlignmentTools(fh, group_name='Alignment_000') as align:
                align.add_alignment_data('template', sam1, sequence1)
                align.add_alignment_data('complement', sam2, sequence2)
                align.add_alignment_data('2d', sam3, sequence1)
        with Fast5File(fname, mode='r') as fh:
            with AlignmentTools(fh, group_name='Alignment_000') as align:
                sam, seq = align.get_alignment_data('template')
                self.assertEqual(sam1, sam)
                self.assertEqual(sequence1, seq)
                sam, seq = align.get_alignment_data('complement')
                self.assertEqual(sam2, sam)
                self.assertEqual(sequence2, seq)
                sam, seq = align.get_alignment_data('2d')
                self.assertEqual(sam3, sam)
                self.assertEqual(sequence1, seq)
                results = align.get_results()
                speed_temp = align.calculate_speed('template')
                speed_comp = align.calculate_speed('complement')
                # Make sure we can calculate speed using only what's in the
                # summary
                summary = fh.get_summary_data('Alignment_000')
                template_summary = summary['genome_mapping_template']
                summary_speed_temp = align.calculate_speed('template',
                                                           template_summary)
        self.assertEqual(250, speed_temp)
        self.assertEqual(250, speed_comp)
        self.assertEqual(speed_temp, summary_speed_temp)
        self.assertDictEqual({'status': 'match found',
                              'direction': 'forward',
                              'ref_name': 'Lambda',
                              'ref_span': (100, 200),
                              'seq_span': (1, 101),
                              'seq_len': 125,
                              'num_aligned': 92,
                              'num_correct': 87,
                              'num_insertions': 8,
                              'num_deletions': 8,
                              'identity': 0.9457,
                              'accuracy': 0.8056}, results['template'])
        self.assertDictEqual({'status': 'match found',
                              'direction': 'reverse',
                              'ref_name': 'Lambda',
                              'ref_span': (100, 200),
                              'seq_span': (0, 96),
                              'seq_len': 120,
                              'num_aligned': 90,
                              'num_correct': 88,
                              'num_insertions': 6,
                              'num_deletions': 10,
                              'identity': 0.9778,
                              'accuracy': 0.8302}, results['complement'])
        self.assertDictEqual({'status': 'match found',
                              'direction': 'forward',
                              'ref_name': 'Lambda',
                              'ref_span': (100, 200),
                              'seq_span': (0, 100),
                              'seq_len': 125,
                              'num_aligned': 98,
                              'num_correct': 96,
                              'num_insertions': 4,
                              'num_deletions': 4,
                              'identity': 0.9796,
                              'accuracy': 0.9057}, results['2d'])



import os
import numpy as np

from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.event_detection import EventDetectionTools
from ont_fast5_api.analysis_tools.segmentation import SegmentationTools
from test.helpers import TestFast5ApiHelper


class TestSegmentationTools(TestFast5ApiHelper):

    def test_001_raw_only(self):
        fname = self.generate_temp_filename()
        with Fast5File(fname, mode='w') as fh:
            fh.add_channel_info({'channel_number': 1,
                                 'sampling_rate': 4000,
                                 'digitisation': 8192,
                                 'range': 819.2,
                                 'offset': 0})
            fh.add_read(12, 'unique_snowflake', 12345, 1000, 0, 120.75)
            raw = np.empty(1000, dtype=np.int16)
            raw[:] = range(1000)
            fh.add_raw_data(raw)
            attrs = {'name': 'test', 'version': 0, 'time_stamp': 'just now'}
            fh.add_analysis('segmentation', 'Segmentation_000', attrs)
            segment_data = {'has_template': 1,
                            'has_complement': 1,
                            'first_sample_template': 10,
                            'duration_template': 470,
                            'first_sample_complement': 520,
                            'duration_complement': 460}
            fh.set_summary_data('Segmentation_000', 'segmentation', segment_data)
            with SegmentationTools(fh, group_name='Segmentation_000') as segment:
                results = segment.get_results()
                self.assertDictEqual({'has_template': True,
                                      'has_complement': True,
                                      'first_sample_template': 10,
                                      'duration_template': 470,
                                      'first_sample_complement': 520,
                                      'duration_complement': 460}, results)
                temp_raw = segment.get_raw_data('template', scale=False)
                np.testing.assert_array_equal(temp_raw, raw[10:480])
                comp_raw = segment.get_raw_data('complement', scale=False)
                np.testing.assert_array_equal(comp_raw, raw[520:980])
                temp_raw, comp_raw = segment.get_raw_data('both', scale=False)
                np.testing.assert_array_equal(temp_raw, raw[10:480])
                np.testing.assert_array_equal(comp_raw, raw[520:980])
                temp_raw, comp_raw = segment.get_raw_data('both', scale=True)
                scaled_temp = raw[10:480] * 0.1
                scaled_comp = raw[520:980] * 0.1
                np.testing.assert_array_almost_equal(temp_raw, scaled_temp, decimal=5)
                np.testing.assert_array_almost_equal(comp_raw, scaled_comp, decimal=5)

    def test_002_events_only(self):
        fname = self.generate_temp_filename()
        with Fast5File(fname, mode='w') as fh:
            fh.add_channel_info({'channel_number': 1,
                                 'sampling_rate': 4000,
                                 'digitisation': 8192,
                                 'range': 819.2,
                                 'offset': 0})
            fh.add_read(12, 'unique_snowflake', 10000, 1000, 0, 120.75)
            with EventDetectionTools(fh, group_name='EventDetection_000', meta={'name': 'test'}) as evdet:
                data = np.zeros(100, dtype=[('start', int), ('length', int), ('mean', float), ('stdv', float)])
                data['start'][2] = 10010
                data['start'][46] = 10470
                data['length'][46] = 10
                data['start'][53] = 10520
                data['start'][97] = 10960
                data['length'][97] = 20
                read_attrs = {'read_number': 12}
                evdet.set_event_data(data, read_attrs)
            attrs = {'name': 'test', 'version': 0, 'time_stamp': 'just now',
                     'event_detection': 'Analyses/EventDetection_000'}
            fh.add_analysis('segmentation', 'Segmentation_000', attrs)
            segment_data = {'has_template': 1,
                            'has_complement': 1,
                            'start_event_template': 2,
                            'end_event_template': 47,
                            'start_event_complement': 53,
                            'end_event_complement': 98}
            fh.set_summary_data('Segmentation_000', 'segmentation', segment_data)
            with SegmentationTools(fh, group_name='Segmentation_000') as segment:
                results = segment.get_results()
                self.assertDictEqual({'has_template': True,
                                      'has_complement': True,
                                      'start_event_template': 2,
                                      'end_event_template': 47,
                                      'start_event_complement': 53,
                                      'end_event_complement': 98,
                                      'first_sample_template': 10,
                                      'duration_template': 470,
                                      'first_sample_complement': 520,
                                      'duration_complement': 460}, results)



__version__ = '4.1.3'
__version_info__ = tuple([int(num) for num in __version__.split('.')])
CURRENT_FAST5_VERSION = 2.0

import sys
if sys.version_info < (3,):
    raise ImportError(
    """ont-fast5-api requires Python 3.7

    Somehow you have ended up running this on Python 2, which reached its end of
    life in 2019. Apologies! To avoid this issue, either:

    - Upgrade to Python 3, or

    - Download an older ont-fast5-api version:

    $ pip install 'ont-fast5-api<3.0'

    Note that you will be missing features and bug fixes by running older versions
    of ont-fast5-api.

    """)

# Set up a default NullHandler in case we don't end up using another one
# Taken from http://docs.python-guide.org/en/latest/writing/logging/
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from ont_fast5_api.compression_settings import register_plugin
register_plugin()



import numpy as np
import os
from numpy import array, ndarray, dtype

from ont_fast5_api.data_sanitisation import _sanitize_data_for_reading, _sanitize_data_for_writing, _clean
from ont_fast5_api.multi_fast5 import MultiFast5File
from test.helpers import TestFast5ApiHelper, test_data


class TestDataSanitisation(TestFast5ApiHelper):

    def test__clean(self):
        self.assertEqual(_clean(1), 1)
        self.assertEqual(_clean(b''), '')
        self.assertEqual(_clean(''), '')
        self.assertEqual(_clean('str'), 'str')
        self.assertEqual(_clean(b'str'), 'str')
        self.assertTrue(isinstance(_clean('str'), str))

        # _clean should convert byte strings into utf-8 ones
        test_str = array(b'Hello!', dtype=bytes)
        self.assertEqual(type(test_str), ndarray)
        self.assertEqual(_clean(test_str), 'Hello!')

        # _clean shouldn't do anything to python strings
        test_str = array('Hello!', dtype=str)
        self.assertEqual(type(test_str), ndarray)
        self.assertEqual(_clean(test_str), test_str)

        self.assertEqual(_clean(array([1, 2, 3])), [1, 2, 3])

    def test__sanitize_data(self):
        # We expect conversion from utf8 to bytestrings and vice-versa
        test_string = 'Avast'
        self.assertEqual(test_string,
                         _sanitize_data_for_reading(test_string.encode()))
        self.assertEqual(test_string.encode(),
                         _sanitize_data_for_writing(test_string))

        test_array = array('Arr', dtype=str)
        self.assertEqual(test_array,
                         _sanitize_data_for_reading(np.char.encode(test_array)))
        self.assertEqual(np.char.encode(test_array),
                         _sanitize_data_for_writing(test_array))

        test_ndarray_utf8 = array([('Narr', 0)],
                                  dtype=[('string', (str, 4)),
                                         ('int', int)])
        test_ndarray_bytes = array([(b'Narr', 0)],
                                   dtype=[('string', (bytes, 4)),
                                          ('int', int)])
        self.assertEqual(test_ndarray_utf8,
                         _sanitize_data_for_reading(test_ndarray_bytes))
        self.assertEqual(test_ndarray_bytes,
                         _sanitize_data_for_writing(test_ndarray_utf8))

    def test__sanitize_data_emptystrings(self):
        test_ndarray_utf8 = array([('', '')], dtype=[('empty', str),
                                                     ('string', str)])
        test_ndarray_bytes = array([('', '')], dtype=[('empty', bytes),
                                                      ('string', bytes)])

        str_array = _sanitize_data_for_reading(test_ndarray_bytes)
        byte_array = _sanitize_data_for_writing(test_ndarray_utf8)
        self.assertTrue(np.array_equal(test_ndarray_utf8, str_array))
        self.assertTrue(np.array_equal(test_ndarray_bytes, byte_array))

    def test_sanitise_array_empty_string(self):
        input_list = [('', 1, 4.8), ('', 2, 7.6)]
        input_types = [('base', str), ('length', 'i4'), ('score', 'f8')]
        input_array = array(input_list, dtype=input_types)
        output_array = _sanitize_data_for_writing(input_array)

        expected_types = [('base', 'S'), ('length', 'i4'), ('score', 'f8')]
        self.assertEqual(expected_types, output_array.dtype)

        roundtrip_array = _sanitize_data_for_reading(output_array)
        self.assertTrue(np.array_equal(input_array, roundtrip_array))

        # Check things works with numpy.recarrays too
        input_rec = input_array.view(np.recarray)
        output_recarray = _sanitize_data_for_writing(input_rec)
        self.assertEqual(expected_types, output_recarray.dtype)

    def test_real_example_file(self):
        with MultiFast5File(os.path.join(test_data, 'rle_basecall_table', 'rle_example.fast5'), 'r') as mf5:
            for read in mf5.get_reads():
                actual_data = read.handle['Analyses/Basecall_1D_000/BaseCalled_template/RunlengthBasecall']
                expected_dtypes = [('base', '<U1'),  # After cleaning this is a unicode string
                                   ('scale', '<f4'),
                                   ('shape', '<f4'),
                                   ('weight', '<f4'),
                                   ('index', '<u4'),
                                   ('runlength', '<u4')]

                for field, expected_type in expected_dtypes:
                    if field != 'base':
                        self.assertEqual(dtype(expected_type), actual_data[field].dtype)
                    else:
                        # Before cleaning the 'base' column is of type byte-string length=1
                        self.assertEqual(dtype('|S1'), actual_data[field].dtype)

                clean_data = _sanitize_data_for_reading(actual_data)
                self.assertEqual(dtype(expected_dtypes), clean_data.dtype)



import numpy as np
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.basecall_2d import Basecall2DTools
from test.helpers import TestFast5ApiHelper


class TestBasecall2DTools(TestFast5ApiHelper):

    def test_001_put_and_retrieve(self):
        fname = self.generate_temp_filename()
        dtypes = [('template', int), ('complement', int)]
        data1 = np.zeros(10, dtype=dtypes)
        data1['template'] = [0, 1, 2, 2, 3, 4, 5, 6, 7, 8]
        data1['complement'] = [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
        dtypes.append(('model_state', '<U5'))
        data2 = np.zeros(10, dtype=dtypes)
        data2['template'] = data1['template']
        data2['complement'] = data1['complement']
        data2['model_state'] = ['AAAAA', 'AAAAT', 'AAATC', 'AAATC', 'ATCCG',
                                'TCCGT', 'CCGTT', 'CGTTA', 'CGTTA', 'GTTAC']
        seq = 'AAAAATCCGTTAC'
        qstring = 'blahblahblahb'
        with Fast5File(fname, mode='w') as fh:
            fh.add_channel_info({'channel_number': 1,
                                 'sampling_rate': 4000,
                                 'digitisation': 8192,
                                 'range': 819.2,
                                 'offset': 0})
            fh.add_read(12, 'unique_snowflake', 12345, 1000, 0, 120.75)
            attrs = {'name': 'test', 'version': 0, 'time_stamp': 'just now'}
            fh.add_analysis('basecall_2d', 'Basecall_2D_000', attrs)
            with Basecall2DTools(fh, group_name='Basecall_2D_000') as basecall:
                basecall.add_prior_alignment(data1)
                basecall.add_2d_call_alignment(data2)
                basecall.add_called_sequence("2D", 'test_2d', seq, qstring)
        with Fast5File(fname, mode='r') as fh:
            with Basecall2DTools(fh, group_name='Basecall_2D_000') as basecall:
                hp_align = basecall.get_prior_alignment()
                np.testing.assert_array_equal(hp_align, data1)
                bc2d = basecall.get_2d_call_alignment()
                np.testing.assert_array_equal(bc2d, data2)
                n, s, q = basecall.get_called_sequence("2D")
                self.assertEqual(n, 'test_2d')
                self.assertEqual(s, seq)
                self.assertEqual(q, qstring)
